from datetime import timedelta
from http import HTTPStatus
from typing import Any, Dict
from unittest.mock import Mock, patch

from django.test import override_settings
from django.urls import reverse
from django.utils.timezone import now

from app_utils.testing import NoSocketsTestCase

from structuretimers.models import ScheduledNotification, Timer
from structuretimers.tasks import send_test_message_to_webhook
from structuretimers.tests.testdata.factory import (
    CitadelTypeFactory,
    DiscordWebhookFactory,
    EveSolarSystemLowSecFactory,
    NotificationRuleFactory,
    TimerFactory,
    UserNoAccessFactory,
    UserWithAccessFactory,
    UserWithCreateFactory,
    UserWithManageFactory,
)

MODELS_PATH = "structuretimers.models"
FORMS_PATH = "structuretimers.forms"
TASKS_PATH = "structuretimers.tasks"

# TODO: Rewrite tests to work also with AA4


@patch(MODELS_PATH + "._task_calc_timer_distances_for_all_staging_systems", Mock())
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestCreateNewTimer(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.add_timer_url = reverse("structuretimers:add")
        cls.timer_list_url = reverse("structuretimers:timer_list")

    def test_user_with_permission_can_add_timer(self):
        # given
        solar_system = EveSolarSystemLowSecFactory()
        structure_type = CitadelTypeFactory()
        form_data = {
            "structure_name": "Timer 4",
            "eve_solar_system_2": [str(solar_system.id)],
            "structure_type_2": [str(structure_type.id)],
            "timer_type": Timer.Type.ANCHORING,
            "days_left": 1,
            "hours_left": 2,
            "minutes_left": 3,
            "objective": Timer.Objective.HOSTILE,
            "visibility": Timer.Visibility.UNRESTRICTED,
            "_save": "Save",
        }
        self.client.force_login(UserWithCreateFactory())

        # when
        response = self.client.post(self.add_timer_url, data=form_data)

        # assert results
        self.assertRedirects(response, self.timer_list_url)
        obj = Timer.objects.get(structure_name="Timer 4")
        self.assertEqual(obj.eve_solar_system, solar_system)
        self.assertEqual(obj.structure_type, structure_type)
        self.assertEqual(obj.timer_type, Timer.Type.ANCHORING)
        timer_date = now() + timedelta(days=1, hours=2, minutes=3)
        self.assertAlmostEqual(obj.date, timer_date, delta=timedelta(seconds=10))

    def test_user_without_permission_can_not_add_timer(self):
        # given
        solar_system = EveSolarSystemLowSecFactory()
        structure_type = CitadelTypeFactory()
        form_data = {
            "structure_name": "Timer 4",
            "eve_solar_system_2": [str(solar_system.id)],
            "structure_type_2": [str(structure_type.id)],
            "timer_type": Timer.Type.ANCHORING,
            "days_left": 1,
            "hours_left": 2,
            "minutes_left": 3,
            "objective": Timer.Objective.HOSTILE,
            "visibility": Timer.Visibility.UNRESTRICTED,
            "_save": "Save",
        }
        self.client.force_login(UserNoAccessFactory())

        # when
        response = self.client.post(self.add_timer_url, data=form_data)

        # then
        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)


@patch(MODELS_PATH + "._task_calc_timer_distances_for_all_staging_systems", Mock())
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestEditTimer(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.timer_list_url = reverse("structuretimers:timer_list")

    def test_user_with_permission_can_open_page(self):
        # given
        user = UserWithCreateFactory()
        timer: Timer = TimerFactory(user=user)
        self.client.force_login(user)

        # when
        response = self.client.get(reverse("structuretimers:edit", args=[timer.pk]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "structuretimers/timer_edit.html")
        self.assertContains(response, timer.eve_solar_system.name)

    def test_user_can_change_his_own_timer(self):
        # given
        user = UserWithCreateFactory()
        timer: Timer = TimerFactory(user=user)
        self.client.force_login(user)

        owner_name = "The Boys"
        form_data = _make_form_data_from_timer(timer) | {
            "owner_name": owner_name,
        }

        # when
        response = self.client.post(
            reverse("structuretimers:edit", args=[timer.pk]), data=form_data
        )

        # then
        self.assertRedirects(response, self.timer_list_url)
        timer.refresh_from_db()
        self.assertEqual(timer.owner_name, owner_name)

    def test_user_without_permission_can_not_open_page(self):
        # given
        user = UserWithCreateFactory()
        timer = TimerFactory()
        self.client.force_login(user)

        # when
        response = self.client.get(reverse("structuretimers:edit", args=[timer.pk]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)

    def test_manager_can_edit_other_timers(self):
        # given
        user = UserWithManageFactory()
        timer: Timer = TimerFactory()
        self.client.force_login(user)

        owner_name = "The Boys"
        form_data = _make_form_data_from_timer(timer) | {
            "owner_name": owner_name,
        }

        # when
        response = self.client.post(
            reverse("structuretimers:edit", args=[timer.pk]), data=form_data
        )

        # then
        self.assertRedirects(response, self.timer_list_url)
        timer.refresh_from_db()
        self.assertEqual(timer.owner_name, owner_name)

    def test_manager_can_not_edit_other_timers_when_corp_restricted(self):
        # given
        timer: Timer = TimerFactory(visibility=Timer.Visibility.CORPORATION)
        user = UserWithManageFactory()
        self.client.force_login(user)

        # when
        response = self.client.get(reverse("structuretimers:edit", args=[timer.pk]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)

    def test_manager_can_not_edit_other_timers_when_opsec(self):
        # setup
        timer = TimerFactory(is_opsec=True)
        user = UserWithManageFactory()
        self.client.force_login(user)

        # when
        response = self.client.get(reverse("structuretimers:edit", args=[timer.pk]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)


def _make_form_data_from_timer(timer: Timer) -> Dict[str, Any]:
    form_data = {
        "date": timer.date.isoformat(),
        "details_image_url": timer.details_image_url or "",
        "details_notes": timer.details_notes or "",
        "eve_alliance": timer.eve_alliance.pk if timer.eve_alliance else "",
        "eve_character": timer.eve_character.pk if timer.eve_character else "",
        "eve_corporation": timer.eve_corporation.pk if timer.eve_corporation else "",
        "eve_solar_system_2": [timer.eve_solar_system.pk],
        "is_important": timer.is_important,
        "is_opsec": timer.is_opsec,
        "location_details": timer.location_details or "",
        "objective": timer.objective,
        "owner_name": timer.owner_name,
        "structure_name": timer.structure_name,
        "structure_type_2": [timer.structure_type.pk],
        "timer_type": timer.timer_type,
        # "user": timer.user.pk if timer.user else "",
        "visibility": timer.visibility,
    }
    return form_data


@patch(MODELS_PATH + "._task_calc_timer_distances_for_all_staging_systems", Mock())
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestDeleteTimer(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.timer_list_url = reverse("structuretimers:timer_list")

    def test_user_can_delete_own_timer(self):
        # given
        user = UserWithCreateFactory()
        timer: Timer = TimerFactory(user=user)
        self.client.force_login(user)

        # when
        response = self.client.get(reverse("structuretimers:delete", args=[timer.pk]))
        self.assertEqual(response.status_code, HTTPStatus.OK)

        response = self.client.post(reverse("structuretimers:delete", args=[timer.pk]))

        # then
        self.assertRedirects(response, self.timer_list_url)
        self.assertFalse(Timer.objects.filter(pk=timer.pk).exists())

    def test_user_can_not_delete_other_timer(self):
        # given
        user = UserWithCreateFactory()
        timer: Timer = TimerFactory()
        self.client.force_login(user)

        # when
        response = self.client.get(reverse("structuretimers:delete", args=[timer.pk]))

        # then
        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)

    def test_manager_can_delete_timer_from_other(self):
        # given
        user = UserWithManageFactory()
        timer: Timer = TimerFactory()
        self.client.force_login(user)

        # when
        response = self.client.get(reverse("structuretimers:delete", args=[timer.pk]))
        self.assertEqual(response.status_code, HTTPStatus.OK)

        response = self.client.post(reverse("structuretimers:delete", args=[timer.pk]))

        # then
        self.assertRedirects(response, self.timer_list_url)
        self.assertFalse(Timer.objects.filter(pk=timer.pk).exists())


"""
@patch(MODELS_PATH+ ".sleep", new=lambda x: x)
@patch(MODELS_PATH+ ".dhooks_lite.Webhook.execute")
class TestSendNotifications(LoadTestDataMixin, TestCase):
    def setUp(self) -> None:
        self.webhook = DiscordWebhookFactory(
            name="Dummy", url="http://www.example.com"
        )
        self.rule = NotificationRule.objects.create(minutes=NotificationRule.MINUTES_0)
        self.rule.webhooks.add(self.webhook)

    def test_normal(self, mock_execute):
        TimerFactory(
            structure_name="Test_1",
            eve_solar_system=self.system_abune,
            structure_type=self.type_raitaru,
            date=now() + timedelta(seconds=2),
        )
        sleep(3)
        self.assertEqual(mock_execute.call_count, 1)
"""


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODELS_PATH + ".sleep", new=lambda x: x)
@patch(TASKS_PATH + ".notify", spec=True)
@patch(MODELS_PATH + ".dhooks_lite.Webhook.execute", spec=True)
class TestSendTestMessageToWebhook(NoSocketsTestCase):
    def test_without_user(self, mock_execute, mock_notify):
        # given
        webhook = DiscordWebhookFactory()

        # when
        send_test_message_to_webhook.delay(webhook_pk=webhook.pk)

        # then
        self.assertEqual(mock_execute.call_count, 1)
        self.assertFalse(mock_notify.called)

    def test_with_user(self, mock_execute, mock_notify):
        # given
        webhook = DiscordWebhookFactory()
        user = UserWithAccessFactory()

        # when
        send_test_message_to_webhook.delay(webhook_pk=webhook.pk, user_pk=user.pk)

        # then
        self.assertEqual(mock_execute.call_count, 1)
        self.assertTrue(mock_notify.called)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODELS_PATH + "._task_calc_timer_distances_for_all_staging_systems", Mock())
class TestTimerSave(NoSocketsTestCase):
    def test_schedule_notifications_for_new_timers_2(self):
        # given
        NotificationRuleFactory()

        # when
        timer = Timer.objects.create(
            date=now() + timedelta(hours=4),
            eve_solar_system=EveSolarSystemLowSecFactory(),
            structure_type=CitadelTypeFactory(),
        )
        # then
        self.assertTrue(ScheduledNotification.objects.filter(timer=timer).exists())



from pathlib import Path

_current_folder = Path(__file__).parent
_FILENAME_EVEUNIVERSE_TESTDATA = "eveuniverse.json"


def test_data_filename():
    return str(_current_folder / _FILENAME_EVEUNIVERSE_TESTDATA)


def test_image_filename():
    return str(_current_folder / "test_image.jpg")



# Generated by Django 3.2.10 on 2022-01-12 20:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("structuretimers", "0002_distances_from_staging"),
    ]

    operations = [
        migrations.AddField(
            model_name="timer",
            name="last_updated_at",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="timer",
            name="date",
            field=models.DateTimeField(
                db_index=True, help_text="Date when this timer happens", null=True
            ),
        ),
        migrations.AlterField(
            model_name="timer",
            name="timer_type",
            field=models.CharField(
                choices=[
                    ("NO", "Unspecified"),
                    ("AR", "Armor"),
                    ("HL", "Hull"),
                    ("FI", "Final"),
                    ("AN", "Anchoring"),
                    ("UA", "Unanchoring"),
                    ("MM", "Moon Mining"),
                    ("PL", "Preliminary"),
                ],
                default="NO",
                max_length=2,
            ),
        ),
    ]



"""Selectors."""

from django.db.models import QuerySet
from eveuniverse.models import EveType

from structuretimers.constants import EveCategoryId, EveGroupId, EveTypeId


def supported_eve_types() -> QuerySet[EveType]:
    """Return query for all supported types."""
    structures = EveType.objects.filter(
        eve_group__eve_category_id=EveCategoryId.STRUCTURE,
        published=True,
    )
    groups_published = EveType.objects.filter(
        eve_group_id__in=[
            EveGroupId.CONTROL_TOWER,
            EveGroupId.MOBILE_DEPOT,
            EveGroupId.MERCENARY_DEN,
        ],
        published=True,
    )
    groups_unpublished = EveType.objects.filter(
        eve_group_id__in=[EveGroupId.PIRATE_FORWARD_OPERATING_BASE]
    )
    types = EveType.objects.filter(
        id__in=[
            EveTypeId.CUSTOMS_OFFICE,
            EveTypeId.ORBITAL_SKYHOOK,
            EveTypeId.IHUB,
            EveTypeId.TCU,
        ]
    )
    qs = (structures | groups_published | groups_unpublished | types).distinct()
    return qs



"""Models."""

# pylint: disable=too-many-lines

import json
from time import sleep
from typing import List, Optional, Tuple

import dhooks_lite
from multiselectfield import MultiSelectField
from multiselectfield.utils import get_max_length
from simple_mq import SimpleMQ

from django.contrib.auth.models import User
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from eveuniverse.helpers import meters_to_ly
from eveuniverse.models import EveRegion, EveSolarSystem, EveType

from allianceauth.eveonline.evelinks import dotlan
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import get_redis_client
from app_utils.datetime import DATETIME_FORMAT
from app_utils.json import JSONDateTimeDecoder, JSONDateTimeEncoder
from app_utils.urls import reverse_absolute, static_file_absolute_url

from structuretimers import __title__
from structuretimers.app_settings import (
    STRUCTURETIMER_NOTIFICATION_SET_AVATAR,
    STRUCTURETIMERS_NOTIFICATIONS_ENABLED,
)
from structuretimers.managers import (
    DistancesFromStagingManager,
    NotificationRuleManager,
    TimerManager,
)

logger = get_extension_logger(__name__)


def default_avatar_url() -> str:
    """avatar url for all messages"""
    return static_file_absolute_url("structuretimers/img/structuretimers_logo.png")


def _task_calc_staging_system():
    from .tasks import calc_staging_system

    return calc_staging_system


def _task_calc_timer_distances_for_all_staging_systems():
    from .tasks import calc_timer_distances_for_all_staging_systems

    return calc_timer_distances_for_all_staging_systems


def _task_schedule_notifications_for_timer():
    from .tasks import schedule_notifications_for_timer

    return schedule_notifications_for_timer


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app and see timers"),
            ("create_timer", "Can create new timers and edit own timers"),
            ("manage_timer", "Can edit and delete any timer"),
            ("opsec_access", "Can create and see opsec timers"),
        )


class DiscordWebhook(models.Model):
    """A Discord webhook"""

    ZKB_KILLMAIL_BASEURL = "https://zkillboard.com/kill/"
    ICON_SIZE = 128

    # delay in seconds between every message sent to Discord
    # this needs to be >= 1 to prevent 429 Too Many Request errors
    SEND_DELAY = 2

    name = models.CharField(
        max_length=64, unique=True, help_text="short name to identify this webhook"
    )
    url = models.CharField(
        max_length=255,
        unique=True,
        help_text=(
            "URL of this webhook, e.g. "
            "https://discordapp.com/api/webhooks/123456/abcdef"
        ),
    )
    notes = models.TextField(
        null=True,
        default=None,
        blank=True,
        help_text="you can add notes about this webhook here if you want",
    )
    is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text="whether notifications are currently sent to this webhook",
    )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        redis_client = get_redis_client()
        self._main_queue = SimpleMQ(redis_client, f"{__title__}_webhook_{self.pk}_main")
        self._error_queue = SimpleMQ(
            redis_client, f"{__title__}_webhook_{self.pk}_errors"
        )

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name='{self.name}')"

    def send_message(
        self,
        content: Optional[str] = None,
        embeds: Optional[List[dhooks_lite.Embed]] = None,
        tts: Optional[bool] = None,
        username: Optional[str] = None,
        avatar_url: Optional[str] = None,
    ) -> int:
        """Adds Discord message to queue for later sending

        Returns updated size of queue
        Raises ValueError if message is incomplete
        """
        if not content and not embeds:
            raise ValueError("Message must have content or embeds to be valid")

        if embeds:
            embeds_list = [obj.asdict() for obj in embeds]
        else:
            embeds_list = None

        message = {}
        if content:
            message["content"] = content
        if embeds_list:
            message["embeds"] = embeds_list
        if tts:
            message["tts"] = tts
        if username:
            message["username"] = username
        if avatar_url:
            message["avatar_url"] = avatar_url

        return self._main_queue.enqueue(json.dumps(message, cls=JSONDateTimeEncoder))

    def send_queued_messages(self) -> int:
        """Send all messages in the queue to this webhook

        Return number of successful sent messages

        Messages that could not be sent are put back into the queue for later retry
        """
        message_count = 0
        while True:
            message_json = self._main_queue.dequeue()
            if message_json:
                message = json.loads(message_json, cls=JSONDateTimeDecoder)
                logger.debug("Sending message to webhook %s", self)
                if self.send_message_to_webhook(message):
                    message_count += 1
                else:
                    self._error_queue.enqueue(message_json)

                sleep(self.SEND_DELAY)

            else:
                break

        while True:
            message_json = self._error_queue.dequeue()
            if message_json:
                self._main_queue.enqueue(message_json)
            else:
                break

        return message_count

    def queue_size(self) -> int:
        """returns current size of the queue"""
        return self._main_queue.size()

    def clear_queue(self) -> int:
        """deletes all messages from the queue. Returns number of cleared messages."""
        counter = 0
        while True:
            y = self._main_queue.dequeue()
            if y is None:
                break
            counter += 1

        return counter

    def send_message_to_webhook(self, message: dict) -> bool:
        """sends message directly to webhook

        returns True if successful, else False
        """
        hook = dhooks_lite.Webhook(url=self.url)
        if message.get("embeds"):
            embeds = [
                dhooks_lite.Embed.from_dict(embed_dict)
                for embed_dict in message.get("embeds", [])
            ]
        else:
            embeds = None

        response = hook.execute(
            content=message.get("content"),
            embeds=embeds,
            username=message.get("username"),
            avatar_url=message.get("avatar_url"),
            wait_for_response=True,
        )
        logger.debug("headers: %s", response.headers)
        logger.debug("status_code: %s", response.status_code)
        logger.debug("content: %s", response.content)
        if response.status_ok:
            return True

        logger.warning(
            "Failed to send message to Discord. HTTP status code: %d, response: %s",
            response.status_code,
            response.content,
        )
        return False

    @classmethod
    def create_discord_link(cls, name: str, url: str) -> str:
        """Returns a link in markdown format for Disord."""
        return f"[{str(name)}]({str(url)})"

    def send_test_message(self, user: User = None) -> Tuple[str, bool]:
        """Sends a test notification to this webhook and returns send report"""
        try:
            user_text = f" sent by **{user}**" if user else ""
            message = {
                "content": f"Test message for webhook **{self.name}**{user_text}",
                "username": __title__,
                "avatar_url": default_avatar_url(),
            }
            success = self.send_message_to_webhook(message)
        except OSError as ex:
            logger.warning(
                "Failed to send test notification to webhook %s: %s",
                self,
                ex,
                exc_info=True,
            )
            return str(ex), False
        return "(no info)", success

    @staticmethod
    def default_username() -> str:
        """avatar username for all messages"""
        return __title__


# pylint: disable=too-many-locals
class Timer(models.Model):
    """A structure timer"""

    # TODO: Old constants needed to maintain compatibility with other apps
    # during transition only. REMOVE as soon as possible.

    TYPE_NONE = "NO"
    TYPE_ARMOR = "AR"
    TYPE_HULL = "HL"
    TYPE_FINAL = "FI"
    TYPE_ANCHORING = "AN"
    TYPE_UNANCHORING = "UA"
    TYPE_MOONMINING = "MM"

    OBJECTIVE_UNDEFINED = "UN"
    OBJECTIVE_HOSTILE = "HO"
    OBJECTIVE_FRIENDLY = "FR"
    OBJECTIVE_NEUTRAL = "NE"

    VISIBILITY_UNRESTRICTED = "UN"
    VISIBILITY_ALLIANCE = "AL"
    VISIBILITY_CORPORATION = "CO"

    class Type(models.TextChoices):
        """A time type."""

        NONE = "NO", _("Unspecified")
        ARMOR = "AR", _("Armor")
        HULL = "HL", _("Hull")
        FINAL = "FI", _("Final")
        ANCHORING = "AN", _("Anchoring")
        UNANCHORING = "UA", _("Unanchoring")
        MOONMINING = "MM", _("Moon Mining")
        THEFT = "TF", _("Theft")
        PRELIMINARY = "PL", _("Preliminary")  # special timer with no date

        @classmethod
        def choices_for_notification_rules(cls) -> List["Timer.Type"]:
            """Subset of choices suitable for creating notification rules."""
            return [choice for choice in cls.choices if choice[0] != cls.PRELIMINARY]

    class Objective(models.TextChoices):
        """A timer objective."""

        UNDEFINED = "UN", _("undefined")
        HOSTILE = "HO", _("hostile")
        FRIENDLY = "FR", _("friendly")
        NEUTRAL = "NE", _("neutral")

    class Visibility(models.TextChoices):
        """A timer visibility."""

        UNRESTRICTED = "UN", _("unrestricted")
        ALLIANCE = "AL", _("Alliance only")
        CORPORATION = "CO", _("Corporation only")

    class SpaceType(models.TextChoices):
        """A timer space type."""

        UNDEFINED = "UN", _("undefined")
        HIGH_SEC = "HS", _("highsec")
        LOW_SEC = "LS", _("lowsec")
        NULL_SEC = "NS", _("nullsec")
        WH_SPACE = "WS", _("wh space")

        @classmethod
        def from_eve_solar_system(cls, eve_solar_system: Optional[EveSolarSystem]):
            """Determine the space type of a solar system and return it."""
            if not eve_solar_system:
                return cls.UNDEFINED
            if eve_solar_system.is_high_sec:
                return cls.HIGH_SEC
            if eve_solar_system.is_low_sec:
                return cls.LOW_SEC
            if eve_solar_system.is_null_sec:
                return cls.NULL_SEC
            if eve_solar_system.is_w_space:
                return cls.WH_SPACE
            raise NotImplementedError(
                f"System with unknown space type: {eve_solar_system}"
            )

    date = models.DateTimeField(
        db_index=True,
        null=True,
        help_text="Date when this timer happens",
    )
    details_image_url = models.CharField(
        max_length=1024,
        default=None,
        blank=True,
        null=True,
        help_text=(
            "URL for details like a screenshot of the structure's fitting, "
            "e.g. https://www.example.com/route/image.jpg"
        ),
    )
    details_notes = models.TextField(
        default="",
        blank=True,
        help_text="Notes with additional information about this timer",
    )
    eve_alliance = models.ForeignKey(
        EveAllianceInfo,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
        help_text="Alliance of the user who created this timer",
    )
    eve_character = models.ForeignKey(
        EveCharacter,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
        help_text="Main character of the user who created this timer",
    )
    eve_corporation = models.ForeignKey(
        EveCorporationInfo,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
        help_text="Corporation of the user who created this timer",
    )
    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        related_name="+",
    )
    is_important = models.BooleanField(
        default=False,
        help_text="Mark this timer as is_important",
    )
    is_opsec = models.BooleanField(
        default=False,
        db_index=True,
        help_text=(
            "Limit access to users with OPSEC clearance. "
            "Can be combined with visibility."
        ),
    )
    location_details = models.CharField(
        max_length=254,
        default="",
        blank=True,
        help_text=(
            "Additional information about the location of this structure, "
            "e.g. name of nearby planet / moon / gate"
        ),
    )
    notification_rules = models.ManyToManyField(
        "NotificationRule",
        through="ScheduledNotification",
        through_fields=("timer", "notification_rule"),
        help_text="Notification rules conforming with this timer",
    )
    objective = models.CharField(
        max_length=2, choices=Objective.choices, default=Objective.UNDEFINED
    )
    owner_name = models.CharField(
        max_length=254,
        default=None,
        blank=True,
        null=True,
        help_text="Name of the corporation owning the structure",
    )
    structure_type = models.ForeignKey(
        EveType, on_delete=models.CASCADE, related_name="+"
    )
    structure_name = models.CharField(max_length=254, default="", blank=True)
    timer_type = models.CharField(max_length=2, choices=Type.choices, default=Type.NONE)
    user = models.ForeignKey(
        User,
        null=True,
        on_delete=models.SET_NULL,
        blank=True,
        related_name="+",
    )
    visibility = models.CharField(
        max_length=2,
        choices=Visibility.choices,
        default=Visibility.UNRESTRICTED,
        db_index=True,
        help_text=(
            "The visibility of this timer can be limited to members"
            " of your organization"
        ),
    )
    last_updated_at = models.DateTimeField(auto_now=True)

    objects = TimerManager()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._original_eve_solar_system_id = self.eve_solar_system_id
        self._original_date = self.date
        self._original_timer_type = self.timer_type

    def __str__(self):
        if self.timer_type != Timer.Type.PRELIMINARY and self.date:
            date = f" @ {self.date.strftime(DATETIME_FORMAT)}"
        else:
            date = ""
        timer_type = self.get_timer_type_display()
        structure_name = self.structure_display_name
        return f"{timer_type} timer for {structure_name}{date}"

    def get_absolute_url(self) -> str:
        """Returns the absolute URL of a timer."""
        url = reverse("structuretimers:timer_list")
        return (
            f"{url}?tab=preliminary"
            if self.timer_type == self.Type.PRELIMINARY
            else url
        )

    @property
    def structure_display_name(self) -> str:
        """Return structure name for display."""
        type_name = self.structure_type.name
        structure_name = f' "{self.structure_name}"' if self.structure_name else ""
        solar_system = self.eve_solar_system.name if self.eve_solar_system else ""
        location = f" near {self.location_details}" if self.location_details else ""
        return f"{type_name}{structure_name} in {solar_system}{location}"

    @property
    def space_type(self) -> "SpaceType":
        """Return space type of a timer."""
        return self.SpaceType.from_eve_solar_system(self.eve_solar_system)

    def user_can_edit(self, user: User) -> bool:
        """Checks if the given user can edit this timer. Returns True or False"""
        return user.has_perm("structuretimers.manage_timer") or (
            self.user == user and user.has_perm("structuretimers.create_timer")
        )

    def label_type_for_timer_type(self) -> str:
        """returns the Boostrap label type for a timer_type"""
        label_types_map = {
            self.Type.NONE: "secondary",
            self.Type.ARMOR: "danger",
            self.Type.HULL: "danger",
            self.Type.FINAL: "danger",
            self.Type.ANCHORING: "warning",
            self.Type.UNANCHORING: "warning",
            self.Type.MOONMINING: "success",
            self.Type.THEFT: "warning",
        }
        if self.timer_type in label_types_map:
            label_type = label_types_map[self.timer_type]
        else:
            label_type = "secondary"
        return label_type

    def label_type_for_objective(self) -> str:
        """returns the Boostrap label type for objective"""
        label_types_map = {
            self.Objective.FRIENDLY: "primary",
            self.Objective.HOSTILE: "danger",
            self.Objective.NEUTRAL: "info",
            self.Objective.UNDEFINED: "secondary",
        }
        if self.objective in label_types_map:
            label_type = label_types_map[self.objective]
        else:
            label_type = "secondary"
        return label_type

    def send_notification(
        self, webhook: DiscordWebhook, content: Optional[str] = None
    ) -> None:
        """Sends notification related to this timer to given webhook."""
        structure_type_name = self.structure_type.name
        solar_system_name = self.eve_solar_system.name if self.eve_solar_system else ""
        title = f"{structure_type_name} in {solar_system_name}"
        if self.structure_name:
            structure_name_text = f'**{structure_type_name}** "{self.structure_name}"'
        else:
            article = "an" if structure_type_name[0:1].lower() in "aeiou" else "a"
            structure_name_text = f"{article} **{structure_type_name}**"

        region_name = (
            self.eve_solar_system.eve_constellation.eve_region.name
            if self.eve_solar_system
            else ""
        )
        solar_system_link = webhook.create_discord_link(
            name=solar_system_name, url=dotlan.solar_system_url(solar_system_name)
        )
        solar_system_text = f"{solar_system_link} ({region_name})"
        near_text = f" near {self.location_details}" if self.location_details else ""
        owned_text = f" owned by **{self.owner_name}**" if self.owner_name else ""
        elapse_at = self.date.strftime(DATETIME_FORMAT) if self.date else "?"
        description = (
            f"The **{self.get_timer_type_display()}** timer for "
            f"{structure_name_text} in {solar_system_text}{near_text}{owned_text} "
            f"will elapse at **{elapse_at}**. "
            f"Our stance is: **{self.get_objective_display()}**."
        )
        structure_icon_url = self.structure_type.icon_url(size=128)
        if self.objective == self.Objective.FRIENDLY:
            color = int("0x375a7f", 16)
        elif self.objective == self.Objective.HOSTILE:
            color = int("0xd9534f", 16)
        else:
            color = None

        embed = dhooks_lite.Embed(
            title=title,
            url=reverse_absolute("structuretimers:timer_list"),
            description=description,
            thumbnail=dhooks_lite.Thumbnail(structure_icon_url),
            color=color,
        )
        if STRUCTURETIMER_NOTIFICATION_SET_AVATAR:
            username = __title__
            avatar_url = default_avatar_url()
        else:
            username = None
            avatar_url = None
        webhook.send_message(
            content=content,
            embeds=[embed],
            username=username,
            avatar_url=avatar_url,
        )


@receiver(post_save, sender=Timer)
def handle_timer_save(
    sender,  # pylint: disable=unused-argument
    instance: Timer,
    created: bool,
    **kwargs,
):
    """Update timer distances from staging and schedule notifications as needed
    after a timer was saved.
    """
    schedule_notifications = (
        STRUCTURETIMERS_NOTIFICATIONS_ENABLED
        and instance.timer_type != Timer.Type.PRELIMINARY
    )
    date_changed = instance.date != instance._original_date
    needs_recalc = (
        created
        or date_changed
        or instance.eve_solar_system.id != instance._original_eve_solar_system_id
    )
    if needs_recalc:
        instance.distances.all().delete()
        _task_calc_timer_distances_for_all_staging_systems().apply_async(
            args=[instance.pk], priority=4
        )
    if (
        instance.timer_type == Timer.Type.PRELIMINARY
        and instance._original_timer_type != Timer.Type.PRELIMINARY
    ):
        instance.scheduled_notifications.all().delete()
    if schedule_notifications and (created or date_changed):
        _task_schedule_notifications_for_timer().apply_async(
            kwargs={"timer_pk": instance.pk, "is_new": created}, priority=3
        )


class NotificationRule(models.Model):
    """A notification rule."""

    class Trigger(models.TextChoices):
        """A trigger choice."""

        NEW_TIMER_CREATED = "TC", _("New timer created")
        SCHEDULED_TIME_REACHED = "TR", _("Scheduled time reached")

    # Minutes choices
    MINUTES_0 = 0
    MINUTES_5 = 5
    MINUTES_10 = 10
    MINUTES_15 = 15
    MINUTES_30 = 30
    MINUTES_45 = 45
    MINUTES_60 = 60
    MINUTES_90 = 90
    MINUTES_120 = 120
    MINUTES_CHOICES = (
        (None, "---------"),
        (MINUTES_0, _("T - 0 minutes")),
        (MINUTES_5, _("T - 5 minutes")),
        (MINUTES_10, _("T - 10 minutes")),
        (MINUTES_15, _("T - 15 minutes")),
        (MINUTES_30, _("T - 30 minutes")),
        (MINUTES_45, _("T - 45 minutes")),
        (MINUTES_60, _("T - 60 minutes")),
        (MINUTES_90, _("T - 90 minutes")),
        (MINUTES_120, _("T - 120 minutes")),
    )

    class PingType(models.TextChoices):
        """Ping type for a notification rule."""

        NONE = "PN", "(no ping)"
        HERE = "PH", "@here"
        EVERYONE = "PE", "@everyone"

        def to_text(self) -> str:
            """returns the text for creating the given ping on Discord"""
            my_map = {self.NONE: "", self.HERE: "@here", self.EVERYONE: "@everyone"}
            try:
                return my_map[self]
            except KeyError:
                return ""

    class Clause(models.TextChoices):
        """A clause in a notification rule."""

        ANY = "AN", "any"
        REQUIRED = "RQ", "required"
        EXCLUDED = "EX", "excluded"

    trigger = models.CharField(
        max_length=2,
        choices=Trigger.choices,
        help_text="Trigger for sending a notification",
    )
    scheduled_time = models.PositiveIntegerField(
        choices=MINUTES_CHOICES,
        null=True,
        default=None,
        blank=True,
        db_index=True,
        help_text=(
            "When to sent a notification in relation to when the timer elapses. "
            "Use together with `Scheduled time reached` trigger."
        ),
    )
    webhook = models.ForeignKey(
        DiscordWebhook,
        on_delete=models.CASCADE,
        help_text="The webhook all notifications are sent to",
    )
    ping_type = models.CharField(
        max_length=2,
        choices=PingType.choices,
        default=PingType.NONE,
        help_text="Options for pinging on notification",
    )
    is_enabled = models.BooleanField(
        default=True,
        help_text="whether this rule is currently active",
    )
    require_timer_types = MultiSelectField(
        choices=Timer.Type.choices_for_notification_rules(),
        max_length=get_max_length(Timer.Type.choices_for_notification_rules(), None),
        blank=True,
        help_text=(
            "Timer must have one of the given timer types "
            "or leave blank to match any."
        ),
    )
    exclude_timer_types = MultiSelectField(
        choices=Timer.Type.choices_for_notification_rules(),
        max_length=get_max_length(Timer.Type.choices_for_notification_rules(), None),
        blank=True,
        help_text="Timer must NOT have one of the given timer types",
    )
    require_objectives = MultiSelectField(
        choices=Timer.Objective.choices,
        max_length=get_max_length(Timer.Objective.choices, None),
        blank=True,
        help_text=(
            "Timer must have one of the given objectives "
            "or leave blank to match any."
        ),
    )
    exclude_objectives = MultiSelectField(
        choices=Timer.Objective.choices,
        max_length=get_max_length(Timer.Objective.choices, None),
        blank=True,
        help_text="Timer must NOT have one of the given objectives",
    )
    require_corporations = models.ManyToManyField(
        EveCorporationInfo,
        blank=True,
        related_name="+",
        help_text=(
            "Timer must be created by one of the given corporations "
            "or leave blank to match any."
        ),
    )
    exclude_corporations = models.ManyToManyField(
        EveCorporationInfo,
        blank=True,
        related_name="notification_rule_exclude_corporations",
        help_text="Timer must NOT be created by one of the given corporations",
    )
    require_alliances = models.ManyToManyField(
        EveAllianceInfo,
        blank=True,
        related_name="+",
        help_text=(
            "Timer must be created by one of the given alliances "
            "or leave blank to match any."
        ),
    )
    exclude_alliances = models.ManyToManyField(
        EveAllianceInfo,
        blank=True,
        related_name="+",
        help_text="Timer must NOT be created by one of the given alliances",
    )
    require_visibility = MultiSelectField(
        choices=Timer.Visibility.choices,
        max_length=get_max_length(Timer.Visibility.choices, None),
        blank=True,
        help_text=(
            "Visibility must be one of the selected or leave blank to match any."
        ),
    )
    exclude_visibility = MultiSelectField(
        choices=Timer.Visibility.choices,
        max_length=get_max_length(Timer.Visibility.choices, None),
        blank=True,
        help_text="Visibility must NOT be one of the selected",
    )
    is_important = models.CharField(
        max_length=2,
        choices=Clause.choices,
        default=Clause.ANY,
        help_text="Wether the timer must be important",
    )
    is_opsec = models.CharField(
        max_length=2,
        choices=Clause.choices,
        default=Clause.ANY,
        help_text="Wether the timer must be OPSEC",
    )
    require_regions = models.ManyToManyField(
        EveRegion,
        blank=True,
        related_name="+",
        help_text=(
            "Timer must be created within one of the given regions "
            "or leave blank to match any."
        ),
    )
    exclude_regions = models.ManyToManyField(
        EveRegion,
        blank=True,
        related_name="+",
        help_text="Timer must NOT be created within one of the given regions",
    )
    require_space_types = MultiSelectField(
        choices=Timer.SpaceType.choices,
        max_length=get_max_length(Timer.SpaceType.choices, None),
        blank=True,
        help_text=(
            "Space type must be one of the selected or leave blank to match any."
        ),
    )
    exclude_space_types = MultiSelectField(
        choices=Timer.SpaceType.choices,
        max_length=get_max_length(Timer.SpaceType.choices, None),
        blank=True,
        help_text="Space Type must NOT be one of the selected",
    )

    objects = NotificationRuleManager()

    def __str__(self) -> str:
        return f"Notification Rule #{self.id}"

    @staticmethod
    def _import_schedule_notifications_for_rule() -> object:
        from .tasks import schedule_notifications_for_rule

        return schedule_notifications_for_rule

    @property
    def ping_type_text(self) -> str:
        """Returns the ping type as string."""
        return self.PingType(self.ping_type).to_text()

    def prepend_ping_text(self, text: str) -> str:
        """prepends ping text to given text and returns it as new text string"""
        return f"{self.ping_type_text} {text}" if self.ping_type_text else text

    # pylint: disable=too-many-branches
    def is_matching_timer(self, timer: "Timer") -> bool:
        """returns True if notification rule is matching the given timer, else False"""
        if timer.date is None:
            return False

        is_matching = True
        if is_matching and self.is_important == self.Clause.REQUIRED:
            is_matching = timer.is_important

        if is_matching and self.is_important == self.Clause.EXCLUDED:
            is_matching = not timer.is_important

        if is_matching and self.is_opsec == self.Clause.REQUIRED:
            is_matching = timer.is_opsec

        if is_matching and self.is_opsec == self.Clause.EXCLUDED:
            is_matching = not timer.is_opsec

        if is_matching and self.require_visibility:
            is_matching = timer.visibility in self.require_visibility

        if is_matching and self.exclude_visibility:
            is_matching = timer.visibility not in self.exclude_visibility

        if is_matching and self.require_timer_types:
            is_matching = timer.timer_type in self.require_timer_types

        if is_matching and self.exclude_timer_types:
            is_matching = timer.timer_type not in self.exclude_timer_types

        if is_matching and self.require_objectives:
            is_matching = timer.objective in self.require_objectives

        if is_matching and self.exclude_objectives:
            is_matching = timer.objective not in self.exclude_objectives

        if is_matching and self.require_corporations.exists():
            is_matching = timer.eve_corporation in self.require_corporations.all()

        if is_matching and self.exclude_corporations.exists():
            is_matching = timer.eve_corporation not in self.exclude_corporations.all()

        if is_matching and self.require_alliances.exists():
            is_matching = timer.eve_alliance in self.require_alliances.all()

        if is_matching and self.exclude_alliances.exists():
            is_matching = timer.eve_alliance not in self.exclude_alliances.all()

        if is_matching and timer.eve_solar_system and self.require_regions.exists():
            is_matching = (
                timer.eve_solar_system.eve_constellation.eve_region
                in self.require_regions.all()
            )

        if is_matching and timer.eve_solar_system and self.exclude_regions.exists():
            is_matching = (
                timer.eve_solar_system.eve_constellation.eve_region
                not in self.exclude_regions.all()
            )
        if is_matching and self.require_space_types:
            is_matching = timer.space_type in self.require_space_types

        if is_matching and self.exclude_space_types:
            is_matching = timer.space_type not in self.exclude_space_types

        return is_matching


@receiver(post_save, sender=NotificationRule)
def handle_rule_save(
    sender,  # pylint: disable=unused-argument
    instance: NotificationRule,
    **kwargs,
):
    """Update scheduled notifications a needed on save."""
    if (
        STRUCTURETIMERS_NOTIFICATIONS_ENABLED
        and instance.is_enabled
        and instance.trigger == NotificationRule.Trigger.SCHEDULED_TIME_REACHED
    ):
        instance._import_schedule_notifications_for_rule().apply_async(
            kwargs={"notification_rule_pk": instance.pk}, priority=4
        )

    if instance.trigger == NotificationRule.Trigger.NEW_TIMER_CREATED:
        instance.scheduled_notifications.all().delete()


class ScheduledNotification(models.Model):
    """A scheduled notification task"""

    timer = models.ForeignKey(
        Timer, on_delete=models.CASCADE, related_name="scheduled_notifications"
    )
    notification_rule = models.ForeignKey(
        NotificationRule,
        on_delete=models.CASCADE,
        related_name="scheduled_notifications",
    )

    timer_date = models.DateTimeField(db_index=True)
    notification_date = models.DateTimeField(db_index=True)
    celery_task_id = models.CharField(max_length=765, default="")

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["timer", "notification_rule"],
                name="unique_notification_schedule",
            )
        ]

    def __repr__(self) -> str:
        return (
            f"ScheduledNotification(timer='{self.timer}', "
            f"notification_rule='{self.notification_rule}', "
            f"celery_task_id='{self.celery_task_id}', "
            f"timer_date='{self.timer_date}', "
            f"notification_date='{self.notification_date}')"
        )


class StagingSystem(models.Model):
    """A staging system."""

    eve_solar_system = models.OneToOneField(
        EveSolarSystem,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )  # TODO: Remove Nullable if possible, because it is causing issues
    is_main = models.BooleanField(default=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._original_eve_solar_system_id = self.eve_solar_system_id

    def __str__(self) -> str:
        return str(self.eve_solar_system)


@receiver(post_save, sender=StagingSystem)
def handle_staging_system_save(
    sender,  # pylint: disable=unused-argument
    instance: StagingSystem,
    created: bool,
    **kwargs,
):
    """Update distances for staging system on save as needed."""
    needs_recalc = (
        created
        or instance.eve_solar_system.id != instance._original_eve_solar_system_id
    )
    if instance.is_main:
        StagingSystem.objects.exclude(pk=instance.pk).update(is_main=False)
    if needs_recalc:
        instance.distances.all().delete()
        _task_calc_staging_system().delay(instance.pk)


class DistancesFromStaging(models.Model):
    """A distance from a staging system."""

    timer = models.ForeignKey(Timer, on_delete=models.CASCADE, related_name="distances")
    staging_system = models.ForeignKey(
        StagingSystem, on_delete=models.CASCADE, related_name="distances"
    )

    light_years = models.FloatField(null=True, default=None, blank=True)
    jumps = models.PositiveIntegerField(null=True, default=None, blank=True)
    updated_at = models.DateTimeField(auto_now=True)

    objects = DistancesFromStagingManager()

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["timer", "staging_system"], name="fpk_distances_from_staging"
            )
        ]

    def __str__(self) -> str:
        return f"{self.timer}-{self.staging_system}"

    def calculate(self):
        """Calculate all distances."""
        if self.staging_system.eve_solar_system:
            self.light_years = meters_to_ly(
                self.staging_system.eve_solar_system.distance_to(
                    self.timer.eve_solar_system
                )
            )
            self.jumps = self.staging_system.eve_solar_system.jumps_to(
                self.timer.eve_solar_system
            )



import datetime as dt
from typing import Generic, TypeVar
from unittest.mock import Mock, patch

import factory
import factory.fuzzy

from django.contrib.auth.models import User
from django.utils.timezone import now
from eveuniverse.models import EveSolarSystem, EveType
from eveuniverse.tests.testdata.factories_2 import (
    EveGroupFactory,
    EveSolarSystemFactory,
    EveTypeFactory,
)

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils
from allianceauth.timerboard.models import Timer as AuthTimer
from app_utils.helpers import random_string
from app_utils.testdata_factories import UserMainFactory

from structuretimers.constants import EveCategoryId, EveGroupId
from structuretimers.models import (
    DiscordWebhook,
    DistancesFromStaging,
    NotificationRule,
    ScheduledNotification,
    StagingSystem,
    Timer,
    post_save,
)

T = TypeVar("T")


class BaseMetaFactory(Generic[T], factory.base.FactoryMetaClass):
    def __call__(cls, *args, **kwargs) -> T:
        return super().__call__(*args, **kwargs)


class UserNoAccessFactory(UserMainFactory):
    pass


class UserWithAccessFactory(UserMainFactory):
    """User with basic rights."""

    permissions__ = [
        "structuretimers.basic_access",
    ]


class UserWithCreateFactory(UserMainFactory):
    """User with basic rights."""

    permissions__ = [
        "structuretimers.basic_access",
        "structuretimers.create_timer",
    ]


class UserWithManageFactory(UserMainFactory):
    """User with manager permission."""

    permissions__ = [
        "structuretimers.basic_access",
        "structuretimers.manage_timer",
    ]


supported_structure_values = set(AuthTimer.Structure.values) - {
    AuthTimer.Structure.OTHER
}


class AuthTimerFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[AuthTimer]
):
    class Meta:
        model = AuthTimer

    corp_timer = False
    details = factory.faker.Faker("sentence")
    eve_character = factory.LazyAttribute(lambda o: o.user.profile.main_character)
    eve_corp = factory.LazyAttribute(
        lambda o: o.user.profile.main_character.corporation
    )
    eve_time = factory.fuzzy.FuzzyDateTime(
        now() + dt.timedelta(days=1), now() + dt.timedelta(days=7)
    )
    important = False
    objective = factory.fuzzy.FuzzyChoice(AuthTimer.Objective.values)
    structure = factory.fuzzy.FuzzyChoice(supported_structure_values)
    system = "Amamake"
    timer_type = factory.fuzzy.FuzzyChoice(AuthTimer.TimerType.values)
    user = factory.SubFactory(UserMainFactory)


class EveSolarSystemNullSecFactory(EveSolarSystemFactory):
    security_status = -1.0


class EveSolarSystemLowSecFactory(EveSolarSystemFactory):
    security_status = 0.3


class EveSolarSystemHighSecFactory(EveSolarSystemFactory):
    security_status = 0.9


class EveSolarSystemWSpaceFactory(EveSolarSystemFactory):
    id = factory.Sequence(lambda n: 31_900_000 + n)
    security_status = -1.0


class CitadelTypeFactory(EveTypeFactory):
    eve_group = factory.SubFactory(
        EveGroupFactory,
        eve_category__id=EveCategoryId.STRUCTURE,
        eve_category__name="Structure",
        id=EveGroupId.CITADEL,
        name="Citadel",
    )


class RefineryTypeFactory(EveTypeFactory):
    eve_group = factory.SubFactory(
        EveGroupFactory,
        eve_category__id=EveCategoryId.STRUCTURE,
        eve_category__name="Structure",
        id=EveGroupId.REFINERY,
        name="Refinery",
    )


class SkyhookTypeFactory(EveTypeFactory):
    eve_group = factory.SubFactory(
        EveGroupFactory,
        eve_category__id=EveCategoryId.ORBITAL,
        eve_category__name="Orbitals",
        id=EveGroupId.SKYHOOK,
        name="Skyhook",
    )
    id = 81080
    name = "Orbital Skyhook"


class DiscordWebhookFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[DiscordWebhook]
):
    class Meta:
        model = DiscordWebhook

    name = factory.Sequence(lambda n: f"discord_webhook_{n}")
    url = factory.LazyAttribute(lambda o: f"{o.name}_url")
    is_enabled = True


@factory.django.mute_signals(post_save)
class NotificationRuleFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[NotificationRule]
):
    class Meta:
        model = NotificationRule

    is_enabled = True
    scheduled_time = NotificationRule.MINUTES_15
    trigger = NotificationRule.Trigger.SCHEDULED_TIME_REACHED
    webhook = factory.SubFactory(DiscordWebhookFactory)


@factory.django.mute_signals(post_save)
class TimerFactory(factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Timer]):
    class Meta:
        model = Timer

    eve_solar_system = factory.SubFactory(EveSolarSystemFactory)
    objective = factory.fuzzy.FuzzyChoice(Timer.Objective.values)
    structure_type = factory.SubFactory(CitadelTypeFactory)
    timer_type = factory.fuzzy.FuzzyChoice([Timer.Type.ARMOR, Timer.Type.HULL])

    @factory.lazy_attribute
    def date(self):
        if self.timer_type == Timer.Type.PRELIMINARY:
            return None
        x = factory.fuzzy.FuzzyDateTime(
            start_dt=now() + dt.timedelta(days=1), end_dt=now() + dt.timedelta(days=7)
        ).fuzz()
        return x


class ScheduledNotificationFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[ScheduledNotification]
):
    class Meta:
        model = ScheduledNotification

    celery_task_id = factory.fuzzy.FuzzyText(length=8)
    notification_date = factory.LazyAttribute(
        lambda o: o.timer_date - dt.timedelta(minutes=15)
    )
    timer_date = factory.fuzzy.FuzzyDateTime(
        start_dt=now() + dt.timedelta(days=1), end_dt=now() + dt.timedelta(days=7)
    )


@factory.django.mute_signals(post_save)
class StagingSystemFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[StagingSystem]
):
    class Meta:
        model = StagingSystem

    eve_solar_system = factory.SubFactory(EveSolarSystemFactory)


class DistancesFromStagingFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[DistancesFromStaging]
):
    class Meta:
        model = DistancesFromStaging

    timer = factory.SubFactory(TimerFactory)
    staging_system = factory.SubFactory(StagingSystemFactory)
    light_years = factory.fuzzy.FuzzyFloat(0.5, 50)
    jumps = factory.fuzzy.FuzzyInteger(1, 50)


# -------- old factories


def add_main_to_user(user: User, character: EveCharacter):
    CharacterOwnership.objects.create(
        user=user, owner_hash="x1" + character.character_name, character=character
    )
    user.profile.main_character = character
    user.profile.save()


def create_user(character: EveCharacter) -> User:
    User.objects.filter(username=character.character_name).delete()
    user = AuthUtils.create_user(character.character_name)
    add_main_to_user(user, character)
    AuthUtils.add_permission_to_user_by_name("structuretimers.basic_access", user)
    user = User.objects.get(pk=user.pk)
    return user


def create_distances_from_staging(
    timer: Timer, staging_system: StagingSystem, **kwargs
) -> DistancesFromStaging:
    params = {
        "timer": timer,
        "staging_system": staging_system,
        "light_years": 1.2,
        "jumps": 3,
    }
    params.update(kwargs)
    return DistancesFromStaging.objects.create(**params)


def create_timer(light_years=None, jumps=None, enabled_notifications=False, **kwargs):
    params = {
        "eve_solar_system": EveSolarSystem.objects.get(id=30004984),
        "structure_type": EveType.objects.get(id=35825),
    }
    if "timer_type" not in kwargs or kwargs["timer_type"] != Timer.Type.PRELIMINARY:
        params["date"] = now() + dt.timedelta(days=3)

    params.update(kwargs)
    with patch(
        "structuretimers.models._task_calc_timer_distances_for_all_staging_systems",
        Mock(),
    ):
        if enabled_notifications:
            timer = Timer.objects.create(**params)
        else:
            with patch(
                "structuretimers.models._task_schedule_notifications_for_timer", Mock()
            ):
                timer = Timer.objects.create(**params)
        if light_years or jumps:
            for staging_system in StagingSystem.objects.all():
                DistancesFromStaging.objects.update_or_create(
                    staging_system=staging_system,
                    timer=timer,
                    defaults={"light_years": light_years, "jumps": jumps},
                )
        return timer


def create_staging_system(light_years=None, jumps=None, **kwargs):
    params = {"eve_solar_system": EveSolarSystem.objects.get(id=30045339)}  # enaluri
    params.update(kwargs)
    with patch("structuretimers.models._task_calc_staging_system", Mock()):
        staging_system = StagingSystem.objects.create(**params)
        if light_years or jumps:
            for timer in Timer.objects.all():
                DistancesFromStaging.objects.update_or_create(
                    staging_system=staging_system,
                    timer=timer,
                    defaults={"light_years": light_years, "jumps": jumps},
                )
        return staging_system


def create_discord_webhook(**kwargs):
    if "name" not in kwargs:
        while True:
            name = f"dummy{random_string(8)}"
            if not DiscordWebhook.objects.filter(name=name).exists():
                break
        kwargs["name"] = name
    if "url" not in kwargs:
        kwargs["url"] = f"https://www.example.com/{kwargs['name']}"
    return DiscordWebhook.objects.create(**kwargs)


def create_notification_rule(schedule_notification=False, **kwargs):
    if "webhook" not in kwargs:
        kwargs["webhook"] = create_discord_webhook()
    if "trigger" not in kwargs:
        kwargs["trigger"] = NotificationRule.Trigger.SCHEDULED_TIME_REACHED
    if "scheduled_time" not in kwargs:
        kwargs["scheduled_time"] = 60
    with patch(
        "structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED",
        schedule_notification,
    ):
        return NotificationRule.objects.create(**kwargs)


def create_scheduled_notification(**kwargs):
    if "timer_date" not in kwargs:
        kwargs["timer_date"] = now() + dt.timedelta(hours=1)
    if "notification_date" not in kwargs:
        kwargs["notification_date"] = now() + dt.timedelta(minutes=45)
    if "celery_task_id" not in kwargs:
        kwargs["celery_task_id"] = random_string(8)
    return ScheduledNotification.objects.create(**kwargs)



# Generated by Django 4.0.10 on 2023-04-18 20:36

import multiselectfield.db.fields

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("structuretimers", "0004_add_space_type_and_region_filters"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notificationrule",
            name="exclude_space_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("UN", "undefined"),
                    ("HS", "highsec"),
                    ("LS", "lowsec"),
                    ("NS", "nullsec"),
                    ("WS", "wh space"),
                ],
                help_text="Space Type must NOT be one of the selected",
                max_length=14,
            ),
        ),
        migrations.AlterField(
            model_name="notificationrule",
            name="require_space_types",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("UN", "undefined"),
                    ("HS", "highsec"),
                    ("LS", "lowsec"),
                    ("NS", "nullsec"),
                    ("WS", "wh space"),
                ],
                help_text="Space type must be one of the selected or leave blank to match any.",
                max_length=14,
            ),
        ),
    ]



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now

from app_utils.testdata_factories import (
    EveAllianceInfoFactory,
    EveCorporationInfoFactory,
)
from app_utils.testing import NoSocketsTestCase

from structuretimers.models import NotificationRule, ScheduledNotification, Timer
from structuretimers.tests.testdata.factory import (
    DiscordWebhookFactory,
    EveSolarSystemHighSecFactory,
    EveSolarSystemLowSecFactory,
    EveSolarSystemNullSecFactory,
    EveSolarSystemWSpaceFactory,
    NotificationRuleFactory,
    ScheduledNotificationFactory,
    TimerFactory,
)

MODULE_PATH = "structuretimers.models"


class TestNotificationRule_IsMatchingTimer(NoSocketsTestCase):
    def test_should_match_when_no_rules_set(self):
        # given
        timer = TimerFactory()
        rule: NotificationRule = NotificationRuleFactory()

        # when/then
        self.assertTrue(rule.is_matching_timer(timer))

    def test_should_match_require_timer_types(self):
        rule: NotificationRule = NotificationRuleFactory(
            require_timer_types=[Timer.Type.ARMOR]
        )
        cases = [
            (Timer.Type.ANCHORING, False),
            (Timer.Type.ARMOR, True),
            (Timer.Type.FINAL, False),
            (Timer.Type.HULL, False),
            (Timer.Type.MOONMINING, False),
            (Timer.Type.NONE, False),
            (Timer.Type.PRELIMINARY, False),
            (Timer.Type.THEFT, False),
        ]

        for tc in cases:
            with self.subTest(type=tc[0]):
                timer: Timer = TimerFactory(timer_type=tc[0])
                self.assertIs(rule.is_matching_timer(timer), tc[1])

    def test_should_match_exclude_timer_types(self):
        rule: NotificationRule = NotificationRuleFactory(
            exclude_timer_types=[Timer.Type.ARMOR]
        )
        cases = [
            (Timer.Type.ANCHORING, True),
            (Timer.Type.ARMOR, False),
            (Timer.Type.FINAL, True),
            (Timer.Type.HULL, True),
            (Timer.Type.MOONMINING, True),
            (Timer.Type.NONE, True),
            (Timer.Type.PRELIMINARY, False),
            (Timer.Type.THEFT, True),
        ]

        for tc in cases:
            with self.subTest(type=tc[0]):
                timer: Timer = TimerFactory(timer_type=tc[0])
                self.assertIs(rule.is_matching_timer(timer), tc[1])

    def test_should_never_match_without_date(self):
        # given
        timer = TimerFactory(date=None)
        rule: NotificationRule = NotificationRuleFactory()

        # when/then
        self.assertFalse(rule.is_matching_timer(timer))

    def test_should_match_require_objectives(self):
        rule: NotificationRule = NotificationRuleFactory(
            require_objectives=[Timer.Objective.HOSTILE]
        )
        cases = [
            (Timer.Objective.FRIENDLY, False),
            (Timer.Objective.HOSTILE, True),
            (Timer.Objective.NEUTRAL, False),
            (Timer.Objective.UNDEFINED, False),
        ]

        for tc in cases:
            with self.subTest(type=tc[0]):
                timer: Timer = TimerFactory(objective=tc[0])
                self.assertIs(rule.is_matching_timer(timer), tc[1])

    def test_should_match_exclude_objectives(self):
        rule: NotificationRule = NotificationRuleFactory(
            exclude_objectives=[Timer.Objective.HOSTILE]
        )
        cases = [
            (Timer.Objective.FRIENDLY, True),
            (Timer.Objective.HOSTILE, False),
            (Timer.Objective.NEUTRAL, True),
            (Timer.Objective.UNDEFINED, True),
        ]

        for tc in cases:
            with self.subTest(type=tc[0]):
                timer: Timer = TimerFactory(objective=tc[0])
                self.assertIs(rule.is_matching_timer(timer), tc[1])

    def test_should_match_require_corporations(self):
        corp = EveCorporationInfoFactory()
        rule: NotificationRule = NotificationRuleFactory()
        rule.require_corporations.add(corp)

        match: Timer = TimerFactory(eve_corporation=corp)
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory()
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_exclude_corporations(self):
        corp = EveCorporationInfoFactory()
        rule: NotificationRule = NotificationRuleFactory()
        rule.exclude_corporations.add(corp)

        match: Timer = TimerFactory()
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory(eve_corporation=corp)
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_require_alliance(self):
        all = EveAllianceInfoFactory()
        rule: NotificationRule = NotificationRuleFactory()
        rule.require_alliances.add(all)

        match: Timer = TimerFactory(eve_alliance=all)
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory()
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_exclude_alliance(self):
        all = EveAllianceInfoFactory()
        rule: NotificationRule = NotificationRuleFactory()
        rule.exclude_alliances.add(all)

        match: Timer = TimerFactory()
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory(eve_alliance=all)
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_require_visibility(self):
        rule: NotificationRule = NotificationRuleFactory(
            require_visibility=[Timer.Visibility.CORPORATION]
        )

        match: Timer = TimerFactory(visibility=Timer.Visibility.CORPORATION)
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory()
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_exclude_visibility(self):
        rule: NotificationRule = NotificationRuleFactory(
            exclude_visibility=[Timer.Visibility.CORPORATION]
        )

        match: Timer = TimerFactory()
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory(visibility=Timer.Visibility.CORPORATION)
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_require_important(self):
        rule: NotificationRule = NotificationRuleFactory(
            is_important=NotificationRule.Clause.REQUIRED
        )

        match: Timer = TimerFactory(is_important=True)
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory()
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_exclude_important(self):
        rule: NotificationRule = NotificationRuleFactory(
            is_important=NotificationRule.Clause.EXCLUDED
        )

        match: Timer = TimerFactory()
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory(is_important=True)
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_require_opsec(self):
        rule: NotificationRule = NotificationRuleFactory(
            is_opsec=NotificationRule.Clause.REQUIRED
        )

        match: Timer = TimerFactory(is_opsec=True)
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory()
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_exclude_opsec(self):
        rule: NotificationRule = NotificationRuleFactory(
            is_opsec=NotificationRule.Clause.EXCLUDED
        )

        match: Timer = TimerFactory()
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory(is_opsec=True)
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_require_regions(self):
        solar_system = EveSolarSystemLowSecFactory()
        rule: NotificationRule = NotificationRuleFactory()
        rule.require_regions.add(solar_system.eve_constellation.eve_region)

        match: Timer = TimerFactory(eve_solar_system=solar_system)
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory()
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_exclude_regions(self):
        solar_system = EveSolarSystemLowSecFactory()
        rule: NotificationRule = NotificationRuleFactory()
        rule.exclude_regions.add(solar_system.eve_constellation.eve_region)

        match: Timer = TimerFactory()
        self.assertTrue(rule.is_matching_timer(match))

        no_match: Timer = TimerFactory(eve_solar_system=solar_system)
        self.assertFalse(rule.is_matching_timer(no_match))

    def test_should_match_require_space_types(self):
        rule = NotificationRuleFactory(require_space_types=[Timer.SpaceType.LOW_SEC])

        cases = [
            ("high sec", EveSolarSystemHighSecFactory(), False),
            ("low sec", EveSolarSystemLowSecFactory(), True),
            ("null sec", EveSolarSystemNullSecFactory(), False),
            ("w space", EveSolarSystemWSpaceFactory(), False),
        ]

        for tc in cases:
            with self.subTest(name=tc[0]):
                timer: Timer = TimerFactory(eve_solar_system=tc[1])
                self.assertIs(rule.is_matching_timer(timer), tc[2])

    def test_should_match_exclude_space_types(self):
        rule = NotificationRuleFactory(exclude_space_types=[Timer.SpaceType.LOW_SEC])

        cases = [
            ("high sec", EveSolarSystemHighSecFactory(), True),
            ("low sec", EveSolarSystemLowSecFactory(), False),
            ("null sec", EveSolarSystemNullSecFactory(), True),
            ("w space", EveSolarSystemWSpaceFactory(), True),
        ]

        for tc in cases:
            with self.subTest(name=tc[0]):
                timer: Timer = TimerFactory(eve_solar_system=tc[1])
                self.assertIs(rule.is_matching_timer(timer), tc[2])


@patch(MODULE_PATH + ".NotificationRule._import_schedule_notifications_for_rule")
class TestNotificationRule_UpdateOnSave(NoSocketsTestCase):
    @patch(MODULE_PATH + ".STRUCTURETIMERS_NOTIFICATIONS_ENABLED", True)
    def test_should_schedule_notifications(self, mock_schedule_notifications):
        rule = NotificationRule(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            webhook=DiscordWebhookFactory(),
        )
        rule.save()
        self.assertTrue(mock_schedule_notifications.called)

    @patch(MODULE_PATH + ".STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    def test_should_not_schedule_when_global_disabled(
        self, mock_schedule_notifications
    ):
        rule = NotificationRule(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            webhook=DiscordWebhookFactory(),
        )
        rule.save()
        self.assertFalse(mock_schedule_notifications.called)

    @patch(MODULE_PATH + ".STRUCTURETIMERS_NOTIFICATIONS_ENABLED", True)
    def test_should_not_schedule_when_rule_disabled(self, mock_schedule_notifications):
        rule = NotificationRule(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            webhook=DiscordWebhookFactory(),
            is_enabled=False,
        )
        rule.save()
        self.assertFalse(mock_schedule_notifications.called)

    @patch(MODULE_PATH + ".STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    def test_created_trigger(self, mock_schedule_notifications):
        """
        when trigger is created
        then delete all scheduled notifications based on same rule
        """
        rule: NotificationRule = NotificationRuleFactory(
            trigger=NotificationRule.Trigger.SCHEDULED_TIME_REACHED,
            scheduled_time=NotificationRule.MINUTES_10,
            webhook=DiscordWebhookFactory(),
        )
        timer: Timer = TimerFactory(
            date=now() + dt.timedelta(hours=4),
        )
        obj = ScheduledNotificationFactory(
            timer=timer,
            notification_rule=rule,
            timer_date=timer.date,
            notification_date=timer.date - dt.timedelta(minutes=10),
        )
        rule.trigger = NotificationRule.Trigger.NEW_TIMER_CREATED
        rule.scheduled_time = None
        rule.save()

        self.assertFalse(ScheduledNotification.objects.filter(pk=obj.pk).exists())
